import unittest

import numpy as np
import pandas as pd

import pandas_ml_utils as pdu
from pandas_ml_utils.model.features_and_labels.features_and_labels_extractor import FeatureTargetLabelExtractor
from pandas_ml_utils.model.fitting.splitting import train_test_split


class TestTrainTestData(unittest.TestCase):

    def test_no_training_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "labelA": [1,2,3,4,5]})

        """when"""
        train_ix, test_ix = train_test_split(df.index, 0)

        """then"""
        np.testing.assert_array_almost_equal(train_ix.values, df.index.values)
        self.assertEqual(0, len(test_ix))

    def test_make_training_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5],
                           "labelA": [1, 2, 3, 4, 5]})

        """when"""
        train_ix, test_ix = train_test_split(df.index, test_size=0.5)

        """then"""
        self.assertEqual(2, len(train_ix))
        self.assertEqual(3, len(test_ix))

    def test_make_youngest_training_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5],
                           "labelA": [1, 2, 3, 4, 5]})

        """when"""
        normal_train, normal_test = train_test_split(df.index, test_size=0.5, seed='youngest')
        lagged_train, lagged_test = train_test_split(
            FeatureTargetLabelExtractor(df, pdu.FeaturesAndLabels(["featureA"], ["labelA"], feature_lags=[0, 1])).features_labels_weights_df[0].index,
            test_size=0.5,
            seed='youngest')

        """then expect a test set"""
        self.assertEqual(3, len(normal_test))
        np.testing.assert_array_equal(normal_test[-3:], np.array([2, 3, 4]))
        self.assertEqual(2, len(lagged_train))
        np.testing.assert_array_equal(lagged_test[-2:], np.array([3, 4]))

    def test_youngest_portion(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        """when"""
        train_ix, test_ix = train_test_split(df.index, test_size=0.6, youngest_size=0.25)

        "then"
        self.assertEqual(6, len(test_ix))
        np.testing.assert_array_equal(test_ix[-2:], np.array([8, 9]))

    def test_lag_smoothing_nan(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelB": [5, 4, 3, 2, 1, 0, 1, 2, 3, None]})

        """when lag smoothing is enabled using shift (which is introducing nan into the data frame)"""
        fl = pdu.FeaturesAndLabels(["featureA"], ["labelB"], feature_lags=[0, 1],
                                   lag_smoothing={1: lambda df: df["featureA"].shift(2)})

        f, l, _ = FeatureTargetLabelExtractor(df, fl).features_labels_weights_df
        len_features = 10 - 1 - 2
        len_none_lables = 1

        """then"""
        self.assertEqual(len(f), len_features - len_none_lables)
        self.assertAlmostEqual(f["featureA", 1].iloc[0], 1.0)
        self.assertAlmostEqual(f["featureA", 1].iloc[-1], 6.0)

    def test_hashable_features_and_labels(self):
        """given"""
        a = pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                  lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                 4: lambda df: df[["featureA"]] * 4})

        b = pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                  lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                 4: lambda df: df[["featureA"]] * 4})

        """expect"""
        self.assertEqual(hash(a), hash(a))
        self.assertEqual(a, a)

        self.assertEqual(a.__id__(), b.__id__())
        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))

    def test_feature_scaling_3d(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureC": [11, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureB": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        """when"""
        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_lags=[1, 2],
                                   feature_rescaling={("featureA", "featureC"): (-1, 1)})

        f = FeatureTargetLabelExtractor(df, fl).features_df.values

        """then"""
        self.assertEqual((8, 2, 3), f.shape)
        self.assertEqual(f[0, 1, 2], 1)
        self.assertEqual(f[0, 1, 0], -1)
        np.testing.assert_array_almost_equal(f[0, :, 1], df["featureB"][[1, 0]])

    def test_feature_scaling_2d(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureC": [11, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureB": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        """when"""
        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_rescaling={("featureA", "featureC"): (-1, 1)})

        f = FeatureTargetLabelExtractor(df, fl).features_df.values

        "then"
        np.testing.assert_array_almost_equal(f[0], np.array([-1, 0.1, 1]))

    def test_lagging(self):
        """given"""
        df = pd.DataFrame({"featureA": [0.5592344 , 0.60739384, 0.19994533, 0.56642537, 0.50965677,
                                        0.168989  , 0.94080671, 0.76651769, 0.8403563 , 0.4003567 ,
                                        0.24295908, 0.50706317, 0.66612371, 0.4020924 , 0.21776017,
                                        0.32559497, 0.12721287, 0.13904584, 0.65887554, 0.08830925],
                           "featureC": [0.43700002, 0.36804634, 0.37568437, 0.34575936, 0.45747071,
                                        0.49749949, 0.09991126, 0.64710179, 0.93479635, 0.40651901,
                                        0.22387503, 0.18458239, 0.0271485 , 0.68792433, 0.69638729,
                                        0.60715125, 0.52557556, 0.88319929, 0.34808869, 0.50250121],
                           "featureB": [0.6784767 , 0.184668  , 0.02049038, 0.92983967, 0.67628553,
                                        0.71373065, 0.30594832, 0.63038278, 0.78284284, 0.84566334,
                                        0.00558188, 0.15819783, 0.09404578, 0.34460875, 0.69161826,
                                        0.41633249, 0.51130681, 0.66703763, 0.74652599, 0.26560367],
                           "labelA": range(20)})

        """when"""
        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_lags=[0,1,2,3,4])

        f = FeatureTargetLabelExtractor(df, fl).features_df.values

        """then"""
        self.assertEqual(len(f), len(df) - 4)
        np.testing.assert_array_equal(f[0,:,0], df["featureA"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(f[-1,:,0], df["featureA"].values[[-1, -2, -3, -4, -5]])
        np.testing.assert_array_equal(f[0,:,1], df["featureB"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(f[-1,:,1], df["featureB"].values[[-1, -2, -3, -4, -5]])
        np.testing.assert_array_equal(f[0,:,2], df["featureC"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(f[-1,:,2], df["featureC"].values[[-1, -2, -3, -4, -5]])

    def test_make_prediction_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        """when"""
        fl = pdu.FeaturesAndLabels(["featureA"], ["labelA"])
        f = FeatureTargetLabelExtractor(df, fl).features_df.values

        """then"""
        self.assertEqual((10, 1), f.shape)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
import pandas as pd
import numpy as np
from pandas._libs.interval import Interval

from pandas_ml_utils.model.features_and_labels.target_encoder import OneHotEncodedDiscrete, OneHotEncodedTargets


class TestEncoders(TestCase):

    def test__one_hot_encoded_targets(self):
        """given"""
        df = pd.DataFrame({"a": [-0.1, 0, 0.1], "b": [0, 1, 2]})
        encoder = OneHotEncodedTargets("a", np.linspace(-0.1, 0.1, 4, endpoint=True))

        """when"""
        encoded = encoder.encode(df)
        decoded = encoder.decode(encoded)

        """then"""
        np.testing.assert_array_almost_equal(encoded.values, np.array([
            [1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]
        ]))
        self.assertEqual(decoded[1], Interval(-0.03333333333333334, 0.033333333333333326, closed='right'))

    def test__one_hot_discrete(self):
        """given"""
        df = pd.DataFrame({"a": [0, 1, 2], "b": [0, 1, 2]})
        encoder = OneHotEncodedDiscrete("a", 3)

        """when"""
        encoded = encoder.encode(df)
        decoded = encoder.decode(encoded)

        """then"""
        np.testing.assert_array_almost_equal(encoded.values, np.array([
            [1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]
        ]))
        np.testing.assert_array_equal(decoded, np.array([0, 1, 2]))

    def test__one_hot_discrete_with_preprocessor(self):
        """given"""
        df = pd.DataFrame({"a": [0, 1, 2], "b": [0, 1, 2]})
        encoder = OneHotEncodedDiscrete("a", 5, pre_processor=lambda x, fact: x * fact)

        """when"""
        encoded = encoder.encode(df, fact=2, foo=12)
        decoded = encoder.decode(encoded)

        """then"""
        np.testing.assert_array_almost_equal(encoded.values, np.array([
            [1., 0., 0., 0., 0.],
            [0., 0., 1., 0., 0.],
            [0., 0., 0., 0., 1.]
        ]))
        np.testing.assert_array_equal(decoded, np.array([0, 2, 4]))


import io
import base64
import inspect
from collections import OrderedDict
from time import perf_counter as pc
from typing import Callable, Dict, Iterable, Any, List

import numpy as np
import pandas as pd


def join_kwargs(*dicts) -> Dict:
    dict = {}
    for d in dicts:
        dict = {**dict, **d}

    return dict


def log_with_time(log_statement: Callable[[], None]):
    log_statement()
    return pc()


def unfold_parameter_space(parameter_space: Dict[str, Iterable], parameters: Dict[str, Any]) -> List[Dict]:
    if len(parameter_space) > 0:
        # more parameters need to be unfolded
        parameter, space = parameter_space.popitem()
        return list(np.array([unfold_parameter_space(parameter_space.copy(), {**parameters, parameter: argument}) for
                              argument in space]).flat)
    else:
        return parameters


def unique_top_level_columns(df: pd.DataFrame):
    return unique(df.columns.get_level_values(0)) if isinstance(df.columns, pd.MultiIndex) else None


def unique(items):
    return list(OrderedDict.fromkeys(items))


def fig_to_png_base64(fig):
    import matplotlib.pyplot as plt
    with io.BytesIO() as f:
        fig.savefig(f, format="png", bbox_inches='tight')
        image = base64.encodebytes(f.getvalue()).decode("utf-8")
        plt.close(fig)
        return image


def one_hot(index: int, number_of_classes: int):
    vec = np.zeros(number_of_classes)

    if index >= 0:
        vec[index] = 1

    return vec


def suitable_kwargs(func, **kwargs):
    suitable_args = inspect.getfullargspec(func).args
    return {arg: kwargs[arg] for arg in kwargs.keys() if arg in suitable_args}


def call_with_suitable_kwargs(func, **kwargs):
    args = suitable_kwargs(func, **kwargs)
    return func(**args)


def call_callable_dynamic_args(func, *args, **kwargs):
    spec = inspect.getfullargspec(func)
    call_args = []

    for i in range(len(spec.args)):
        if i < len(args):
            call_args.append(args[i])
        elif spec.args[i] in kwargs:
            call_args.append(kwargs[spec.args[i]])
            del kwargs[spec.args[i]]

    # inject eventually still missing var args
    if spec.varargs and len(args) > len(spec.args) and len(args) > len(call_args):
        call_args += args[len(call_args):]

    # inject kwargs if we have some left overs
    if spec.varkw:
        return func(*call_args, **kwargs)
    else:
        return func(*call_args)


def integrate_nested_arrays(arr: np.ndarray) -> np.ndarray:
    if arr is not None and len(arr) > 0 and arr[-1].dtype == 'object':
        if len(arr.shape) > 1 and arr.shape[1] > 1:
            return np.array([[np.array(c) for c in r] for r in arr])
        else:
            return np.array([np.array(c) for r in arr for c in r])
    else:
        return arr


import unittest
import logging
import pandas as pd
import pytest

import pandas_ml_utils.extern.cryptocompare as cc
from pandas_ml_utils.extern.cryptocompare import TIME
from pandas_ml_utils.datafetching.fetch_cryptocompare import _data_to_frame

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
DEBUG = False


class TestCryptoCompare(unittest.TestCase):

    @pytest.mark.skipif(DEBUG is False, reason="test only if needed due to rate limit")
    def test_download_day(self):
        """when"""
        data = cc.get_historical_price_day("BTC", limit=None)["Data"]
        df = _data_to_frame(data)

        """then"""
        print(f"\n{len(data)}")
        self.assertTrue(len(data) > 2000)
        self.assertEqual(len(data), len(df))
        self.assertEqual(pd.DataFrame(data).duplicated("time").astype(int).sum(), 0)

    @pytest.mark.skipif(DEBUG is False, reason="test only if needed due to rate limit")
    def test_bach_load_hour(self):
        data = cc.get_historical_price_hour("BTC", limit=None)["Data"]
        youngest_ts = min([r[TIME] for r in data])

        """then"""
        print(f"\n{len(data)}, {youngest_ts}")
        self.assertTrue(youngest_ts >= 1270810800 - 60 * 60 * 24)
        self.assertTrue(len(data) > 86000)
        self.assertEqual(pd.DataFrame(data).duplicated("time").astype(int).sum(), 0)



from unittest import TestCase

import numpy as np
import pandas as pd

import pandas_ml_utils as pml

print(pml.__version__)


class TestDfExtensions(TestCase):

    def test__inner_join(self):
        """given"""
        dfa = pd.DataFrame({"a": [1, 2, 3]}, index=[0, 1, 2])
        dfb = pd.DataFrame({"a": [1, 2, 3]}, index=[1, 2, 3])

        """when"""
        df = dfa.inner_join(dfb, prefix='b_')

        """then"""
        pd.testing.assert_frame_equal(df, pd.DataFrame({"a": [2, 3], "b_a": [1, 2]}, index=[1, 2]))

    def test__inner_join_multi_index(self):
        """given"""
        dfa = pd.DataFrame({"a": [1, 2, 3]}, index=[0, 1, 2])
        dfa.columns = pd.MultiIndex.from_tuples([("a", "a")])

        dfb = pd.DataFrame({"a": [1, 2, 3]}, index=[1, 2, 3])

        """when"""
        df = dfa.inner_join(dfb, prefix='b')

        """then"""
        res = pd.DataFrame({"a": [2, 3], "b_a": [1, 2]}, index=[1, 2])
        res.columns = pd.MultiIndex.from_tuples([("a", "a"), ("b", "a")])
        pd.testing.assert_frame_equal(df, res)

    def test__cloc2(self):
        """given"""
        df = pd.DataFrame({"a": [2, 3], "b_a": [1, 2]}, index=[1, 2])
        df.columns = pd.MultiIndex.from_tuples([("a", "a"), ("b", "a")])

        """when"""
        single = df.cloc2("a")
        multi = df.cloc2(["a"])

        """then"""
        self.assertListEqual(single.columns.tolist(), ["a", "b"])
        self.assertListEqual(multi.columns.tolist(), [("a", "a"), ("a", "b")])
        self.assertIsInstance(multi.columns, pd.MultiIndex)




import logging
import unittest

import pandas as pd
from test.config import TEST_FILE

from pandas_ml_utils.analysis.correlation_analysis import _sort_correlation

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class ComponentTest(unittest.TestCase):

    def test_correlation_matrix(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date')[["spy_Open", "spy_High", "spy_Low", "spy_Close"]]

        """when"""
        corr = _sort_correlation(df.corr(), recursive=True)

        """then"""
        self.assertListEqual(corr.columns.tolist(), ["spy_Open", "spy_High", "spy_Close", "spy_Low"])




from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ml_utils.model.features_and_labels.features_and_labels import FeaturesAndLabels
from pandas_ml_utils.model.features_and_labels.features_and_labels_extractor import FeatureTargetLabelExtractor

DF = pd.DataFrame({"a": [1,2,3,4,5],
                   "b": [5,4,3,2,1],
                   "c": [1,2,3,4,5],
                   "d": [5,4,3,2,1],
                   "e": [5,4,3,2,1],})


class TestFeaturesAndLabelsExtraction(TestCase):

    def test_magic_arguments(self):
        """given"""
        labels = []

        def test(df, _labels):
            labels.append(_labels)
            return df

        """when"""
        FeatureTargetLabelExtractor(pd.DataFrame({}), FeaturesAndLabels(["a"], ["b"], pre_processor=test))

        """then"""
        self.assertListEqual(labels, [["b"]])

    def test_simple(self):
        """given"""
        fl = FeaturesAndLabels(["a"], ["d", "e"],
                               feature_lags=range(3),
                               targets={"b": None})

        """when"""
        f, l, _ = FeatureTargetLabelExtractor(DF, fl).features_labels_weights_df

        """then"""
        np.testing.assert_array_almost_equal(f.values, np.array([[[3], [2], [1]],
                                                                 [[4], [3], [2]],
                                                                 [[5], [4], [3]]]))

    def test_pre_processor(self):
        """given"""
        fl = FeaturesAndLabels(["lala"], ["b"], pre_processor=lambda df, **kwargs: df.rename(columns=kwargs), a="lala")

        """when"""
        f, l, w = FeatureTargetLabelExtractor(DF, fl).features_labels_weights_df

        """then"""
        self.assertEqual(len(f), len(l))
        self.assertListEqual(f.columns.tolist(), ['lala'])
        self.assertListEqual(l.columns.tolist(), ['b'])

    def test_wither(self):
        """given"""
        fl = FeaturesAndLabels(["lala"], ["b"],
                               feature_lags=[0, 1],
                               pre_processor=lambda df, names: df.rename(columns=names), a="lala")

        """when"""
        fl = fl.with_kwargs(a="lolo", b="lala")

        """then"""
        self.assertListEqual(fl.feature_lags, [0, 1])
        self.assertEqual(fl.kwargs["a"], "lolo")
        self.assertEqual(fl.kwargs["b"], "lala")




import logging
import unittest

import pandas as pd
from sklearn.neural_network import MLPClassifier

import pandas_ml_utils as pdu
from pandas_ml_utils.constants import *
from test.config import TEST_FILE

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class PreprocessorTest(unittest.TestCase):

    def test_pre_processor(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date')

        """when"""
        fit = df.fit(
            pdu.SkModel(
                MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001, random_state=42),
                pdu.FeaturesAndLabels(features=['feature'], labels=['label'], label_type=int,
                                      gross_loss=lambda df: df["spy_Close"] - df["spy_Open"],
                                      pre_processor=lambda _df: pdu.LazyDataFrame(
                                          _df,
                                          feature=lambda f: f["vix_Close"].rolling(2).mean(),
                                          label=lambda f: (f["spy_Close"].shift(1) > f["spy_Open"]).shift(-1)).to_dataframe())),
            test_size=0.4,
            test_validate_split_seed=42)

        bt = df.backtest(fit.model)
        p = df.predict(fit.model, 2)

        """then fit"""
        self.assertListEqual(fit.test_summary.df.columns.tolist(), [(PREDICTION_COLUMN_NAME, 'label'), (LABEL_COLUMN_NAME, 'label'), (GROSS_LOSS_COLUMN_NAME, GROSS_LOSS_COLUMN_NAME)])

        """ and backtest"""
        self.assertListEqual(bt.df.columns.tolist(), [(PREDICTION_COLUMN_NAME, 'label'), (LABEL_COLUMN_NAME, 'label'), (GROSS_LOSS_COLUMN_NAME, GROSS_LOSS_COLUMN_NAME), *[(SOURCE_COLUMN_NAME, c) for c in df.columns], (SOURCE_COLUMN_NAME, "feature"), (SOURCE_COLUMN_NAME, "label")])
        self.assertEqual(bt.df.index[-1], "2019-09-13")

        """ and prediction"""
        self.assertListEqual(p.columns.tolist(), [(PREDICTION_COLUMN_NAME, 'label')])
        self.assertEqual(p.index[-1], "2019-09-16")
        self.assertEqual(len(p), 2)



import unittest

import pandas as pd
from copy import deepcopy
from pandas_ml_utils.wrappers.lazy_dataframe import LazyDataFrame


class TestLazyDataFrame(unittest.TestCase):

    def test_lazy_dataframe(self):
        df = pd.DataFrame({"test": [1,2,3,4,5]})
        df2 = pd.DataFrame({"test2": [1,2,3,4,5]})
        ldf = LazyDataFrame(df, foo=lambda _df: _df[_df.columns[0]] * 2)
        ldf2 = ldf.with_dataframe(df2)

        # test lazy functions
        pd.testing.assert_series_equal(ldf["foo"], df["test"] * 2, check_names=False)
        self.assertEqual(ldf[["foo", "test"]].shape, (5,2))
        pd.testing.assert_frame_equal(ldf[["test", "foo"]], ldf.to_dataframe())
        pd.testing.assert_frame_equal(ldf2[["test2", "foo"]], ldf2.to_dataframe())
        pd.testing.assert_frame_equal(ldf[["foo"]], ldf2[["foo"]])

        # test delegation
        pd.testing.assert_index_equal(ldf.index, df.index)
        self.assertEqual(len(ldf.min().index), 2)

        # test structure
        self.assertFalse("foo" in ldf.df)
        self.assertTrue("foo" in ldf)

    def test_deepcopy(self):
        """given"""
        ldf = LazyDataFrame(None, foo=lambda _df: _df["lala"] * 2)

        """when"""
        copy = deepcopy(ldf)

        """then"""
        self.assertListEqual(list(copy.kwargs.keys()), list(ldf.kwargs.keys()))
        self.assertNotEqual(copy.hash, ldf.hash)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ml_utils.utils.functions import call_callable_dynamic_args, integrate_nested_arrays


class TestUtilFunctions(TestCase):

    def test_call_dynamic_args(self):
        """when"""
        arguments = list(range(4))
        l1 = lambda a: a
        l2 = lambda a, b: a + b
        l3 = lambda a, *b: a + sum(b)
        l4 = lambda *a: sum(a)
        def f1(a, b, c, d):
            return a + b + c + d
        def f2(a, *b, **kwargs):
            return a + sum(b)
        def f3(a, b):
            return a + b

        """then"""
        self.assertEqual(call_callable_dynamic_args(l1, *arguments), 0)
        self.assertEqual(call_callable_dynamic_args(l2, *arguments), 1)
        self.assertEqual(call_callable_dynamic_args(l3, *arguments), 6)
        self.assertEqual(call_callable_dynamic_args(l4, *arguments), 6)
        self.assertEqual(call_callable_dynamic_args(f1, *arguments), 6)
        self.assertEqual(call_callable_dynamic_args(f2, *arguments), 6)
        self.assertEqual(call_callable_dynamic_args(f3, *arguments), 1)

    def test_call_dynamic_args_kwargs(self):
        """expect"""
        self.assertTrue(call_callable_dynamic_args(lambda a, b: True, 1, b=2))
        self.assertTrue(call_callable_dynamic_args(lambda a, b: True, a=1, b=2))
        self.assertTrue(call_callable_dynamic_args(lambda a, b: True, 1, 2))
        self.assertRaises(Exception, lambda: call_callable_dynamic_args(lambda a, b: True, 1))
        self.assertRaises(Exception, lambda: call_callable_dynamic_args(lambda a, b: True, 1, c=1))

    def test_inegrate_nested_array(self):
        """given"""
        x = np.array([1, 2])
        df = pd.DataFrame({"a": [np.zeros((4, 3)) for _ in range(10)],
                           "b": [np.ones((4, 3)) for _ in range(10)]})

        """when"""
        res1 = integrate_nested_arrays(df[["a"]].values)
        res2 = integrate_nested_arrays(df.values)
        res3 = integrate_nested_arrays(x)

        """then"""
        self.assertTrue(df.values[-1].dtype == 'object')
        self.assertEqual(res1.shape, (10, 4, 3))
        self.assertEqual(res2.shape, (10, 2, 4, 3))
        self.assertTrue(x is res3)


import numpy as np

from pandas_ml_utils import Model


class MockModel(Model):

    def __init__(self, fnl):
        super().__init__(fnl)
        self.label_shape = None

    def fit(self, x: np.ndarray, y: np.ndarray, x_val: np.ndarray, y_val: np.ndarray, sample_weight_train: np.ndarray,
            sample_weight_test: np.ndarray) -> float:
        self.label_shape = y.shape
        return 0.0

    def predict(self, x: np.ndarray) -> np.ndarray:
        return np.ones((len(x), *self.label_shape[1:]))


import logging
import unittest

import pandas as pd
from sklearn.neural_network import MLPClassifier

import pandas_ml_utils as pdu
from test.config import TEST_FILE

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class LazyDataFrameTest(unittest.TestCase):

    def test_fit_and_co(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date').tail(100)
        ldf = pdu.LazyDataFrame(
            df,
            sma=lambda f: f["vix_Close"].rolling(2).mean(),
            label=lambda f: f["spy_Close"] > f["spy_Open"]
        )
        model = pdu.SkModel(
            MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), random_state=42),
            pdu.FeaturesAndLabels(["sma"], ["label"])
        )

        """when"""
        fit = ldf.fit(model)
        bt = ldf.backtest(fit.model)
        p = ldf.predict(fit.model)

        """then"""
        self.assertEqual(len(fit.test_summary.df), 40)
        self.assertEqual(len(bt.df), 100 - 1)
        self.assertEqual(len(p), 100 - 1)

    def test_pre_process_and_fit_and_co(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date').tail(100)
        model = pdu.SkModel(
            MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), random_state=42),
            pdu.FeaturesAndLabels(
                ["sma"],
                ["label"],
                pre_processor=lambda _df: pdu.LazyDataFrame(
                    _df,
                    sma=lambda f: f["vix_Close"].rolling(2).mean(),
                    label=lambda f: f["spy_Close"] > f["spy_Open"]
                )
            )
        )

        """when"""
        fit = df.fit(model)
        bt = df.backtest(fit.model)
        p = df.predict(fit.model)

        """then"""
        self.assertEqual(len(fit.test_summary.df), 40)
        self.assertEqual(len(bt.df), 100 - 1)
        self.assertEqual(len(p), 100 - 1)




import uuid
from functools import lru_cache
from typing import Callable, Union

import pandas as pd

from pandas_ml_utils.model.fitting.fitter import fit, predict, backtest


class LazyDataFrame(object):

    def __init__(self, df: pd.DataFrame, **kwargs: Callable[[pd.DataFrame], Union[pd.DataFrame, pd.Series]]) -> None:
        self.hash = uuid.uuid4()
        self.df: pd.DataFrame = df
        self.kwargs = kwargs

    @property
    def columns(self):
        return self.df.columns.tolist() + list(self.kwargs.keys())

    @property
    def index(self):
        return self.df.index

    def fit(self, *args, **kwargs):
        return fit(self, *args, **kwargs)

    def backtest(self, *args, **kwargs):
        return backtest(self, *args, **kwargs)

    def predict(self, *args, **kwargs):
        return predict(self, *args, **kwargs)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, item: str):
        return self.to_dataframe()[item]

    def __setitem__(self, key: str, value: Callable[[pd.DataFrame], Union[pd.DataFrame, pd.Series]]):
        self.hash = uuid.uuid4()
        if callable(value):
            self.kwargs[key] = value(self.df)
        else:
            self.df[key] = value

    def __getattr__(self, item):
        if isinstance(item, str) and item.startswith("__") and item.endswith("__"):
            # do not allow private private items
            raise AttributeError(f'{type(self)} has not attr {item}')
        else:
            return self.to_dataframe().__getattr__(item)

    def __contains__(self, key):
        return key in self.to_dataframe()

    def __hash__(self):
        return int(self.hash)

    def __eq__(self, other):
        return self.hash == other.hash if isinstance(other, LazyDataFrame) else False

    def __str__(self):
        return f'{self.hash}, {self.kwargs.keys()}'

    def __deepcopy__(self, memodict={}):
        return LazyDataFrame(self.df, **self.kwargs)

    def with_dataframe(self, df: pd.DataFrame):
        return LazyDataFrame(df, **self.kwargs)

    @lru_cache(maxsize=1)
    def to_dataframe(self) -> pd.DataFrame:
        df = self.df.copy()
        for key, calculation in self.kwargs.items():
            column = calculation(df)
            if isinstance(column, pd.DataFrame):
                df = df.join(column.add_prefix(f'{key}_'))
            else:
                df[key] = column

        return df


import logging
from functools import lru_cache
from time import perf_counter as pc
from typing import Tuple, Dict, Union, List

import numpy as np
import pandas as pd
from sortedcontainers import SortedDict

from pandas_ml_utils.constants import *
from pandas_ml_utils.model.features_and_labels.features_and_labels import FeaturesAndLabels
from pandas_ml_utils.model.features_and_labels.target_encoder import TargetLabelEncoder, \
    MultipleTargetEncodingWrapper, IdentityEncoder
from pandas_ml_utils.model.fitting.splitting import train_test_split
from pandas_ml_utils.utils.classes import ReScaler
from pandas_ml_utils.utils.functions import log_with_time, call_callable_dynamic_args, unique_top_level_columns, \
    join_kwargs, integrate_nested_arrays

_log = logging.getLogger(__name__)


class FeatureTargetLabelExtractor(object):

    def __init__(self, df: pd.DataFrame, features_and_labels: FeaturesAndLabels, **kwargs):
        # prepare fields
        labels = features_and_labels.labels
        encoder = lambda frame, **kwargs: frame
        label_columns = None
        joined_kwargs = join_kwargs(features_and_labels.kwargs, kwargs)

        # eventually transform callable labels to its expected structure
        if callable(labels):
            labels = call_callable_dynamic_args(labels, df, **joined_kwargs)

        # unfold labels, currently supported types are:
        #  Union[List[str], TargetLabelEncoder, Dict[str, Union[List[str], TargetLabelEncoder]]]
        if isinstance(labels, list):
            label_columns = labels
        elif isinstance(labels, TargetLabelEncoder):
            encoder = labels.encode
            label_columns = labels.labels_source_columns
        elif isinstance(labels, Dict):
            # this is our multi model case, here we add an extra dimension to the labels array
            label_columns = [l for ls in labels.values()
                             for l in (ls if isinstance(ls, list) else ls.labels_source_columns)]

            # we need a special encoder which is wrapping all encoder for each target
            encoder = MultipleTargetEncodingWrapper({
                t: l if isinstance(l, TargetLabelEncoder) else IdentityEncoder(l) for t, l in labels.items()
            }).encode

        # assign all fields
        self._features_and_labels = features_and_labels # depricated copy all fields here
        self._features = features_and_labels.features
        self._labels_columns = label_columns
        self._labels = labels
        self._label_type = features_and_labels.label_type
        self._targets = features_and_labels.targets
        self._gross_loss = features_and_labels.gross_loss
        self._encoder = encoder
        self._joined_kwargs = joined_kwargs

        # pre assign this variable
        # but notice that it get overwritten by an engineered data frame later on
        self._df = df

        # this function uses clojures
        def call_dynamic(func, *args):
            joined_kwargs = join_kwargs(self.__dict__, self._joined_kwargs)
            return call_callable_dynamic_args(func, *args, **joined_kwargs)

        self._df = call_dynamic(features_and_labels.pre_processor, df)
        self.__call_dynamic = call_dynamic

    @property
    def df(self):
        return self._df

    @property
    def min_required_samples(self):
        return len(self._df) - len(self.features_df) + 1

    def prediction_to_frame(self,
                            prediction: np.ndarray,
                            index: pd.Index = None,
                            inclusive_labels: bool = False,
                            inclusive_source: bool = False) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        # sanity check
        if not isinstance(prediction, np.ndarray):
            raise ValueError(f"got unexpected prediction: {type(prediction)}\n{prediction}")

        # assign index
        index = self._df.index if index is None else index

        # eventually fix the shape of the prediction
        if len(prediction.shape) == 1:
            prediction = prediction.reshape(len(prediction), 1)

        # prediction_columns
        columns = pd.MultiIndex.from_tuples(self.label_names(PREDICTION_COLUMN_NAME))
        multi_dimension_prediction = len(prediction.shape) > 1 and len(columns) < prediction.shape[1]
        if multi_dimension_prediction:
            if len(prediction.shape) < 3:
                df = pd.DataFrame({"a":[ r.tolist() for r in prediction]}, index=index)
            else:
                df = pd.DataFrame({col: [row.tolist() for row in prediction[:, col]] for col in range(prediction.shape[1])},index=index)

            df.columns = columns
        else:
             df = pd.DataFrame(prediction, index=index, columns=columns)

        # add labels if requested
        if inclusive_labels:
            dfl = self.labels_df
            dfl.columns = pd.MultiIndex.from_tuples(self.label_names(LABEL_COLUMN_NAME))
            df = df.join(dfl, how='inner')

            # add loss if provided
            loss_df = self.gross_loss_df
            df = df.join(loss_df.loc[df.index], how='inner') if loss_df is not None else df

        # add target if provided
        target_df = self.target_df
        df = df.join(target_df.loc[df.index], how='inner') if target_df is not None else df

        # also add source if requested
        if inclusive_source:
            df = df.join(self.source_df, how='inner')

        # finally we can return our nice and shiny df
        return df

    def training_and_test_data(self,
                               test_size: float = 0.4,
                               youngest_size: float = None,
                               seed: int = 42) -> Tuple[Tuple[np.ndarray,...], Tuple[np.ndarray,...]]:
        features, labels, weights = self.features_labels_weights_df
        train_ix, test_ix = train_test_split(features.index, test_size, youngest_size, seed=seed)

        return (
            (train_ix,
             features.loc[train_ix].values,
             integrate_nested_arrays(labels.loc[train_ix].values),
             weights.loc[train_ix].values if weights is not None else None),
            (test_ix,
             features.loc[test_ix].values,
             integrate_nested_arrays(labels.loc[test_ix].values),
             weights.loc[test_ix].values if weights is not None else None)
        )

    @property
    def features_labels_weights_df(self) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        # engineer features and labels
        df_features = self.features_df
        df_labels = self.labels_df
        index_intersect = df_features.index.intersection(df_labels.index)

        # select only joining index values
        df_features = df_features.loc[index_intersect]
        df_labels = df_labels.loc[index_intersect]
        # TODO add proper label weights
        df_weights = None #pd.DataFrame(np.ones(len(df_labels)), index=df_labels.index)

        # sanity check
        if not len(df_features) == len(df_labels):
            raise ValueError(f"unbalanced length of features and labels {len(df_features), len(df_labels)}")

        return df_features, df_labels, df_weights

    @property
    @lru_cache(maxsize=1)
    def features_df(self) -> pd.DataFrame:
        start_pc = log_with_time(lambda: _log.debug(" make features ..."))
        feature_lags = self._features_and_labels.feature_lags
        features = self._features
        lag_smoothing = self._features_and_labels.lag_smoothing
        feature_rescaling = self._features_and_labels.feature_rescaling

        # drop nan's and copy frame
        df = self._df[features].dropna().copy()

        # generate feature matrix
        if feature_lags is None:
            dff = df
        else:
            dff = pd.DataFrame({}, index=df.index)
            # return RNN shaped 3D arrays
            for feature in features:
                feature_series = df[feature]
                smoothers = None

                # smooth out feature if requested
                if lag_smoothing is not None:
                    smoothers = SortedDict({lag: smoother(feature_series.to_frame())
                                            for lag, smoother in lag_smoothing.items()})

                for lag in feature_lags:
                    # if smoothed values are applicable use smoothed values
                    if smoothers is not None and len(smoothers) > 0 and smoothers.peekitem(0)[0] <= lag:
                        feature_series = smoothers.popitem(0)[1]

                    # assign the lagged (eventually smoothed) feature to the features frame
                    dff[(feature, lag)] = feature_series.shift(lag)

            # fix tuple column index to actually be a multi index
            dff.columns = pd.MultiIndex.from_tuples(dff.columns)

            # drop all rows which got nan now
            dff = dff.dropna()

        # do rescaling
        if feature_rescaling is not None:
            for rescale_features, target_range in feature_rescaling.items():
                # tuple need to be converted to list!
                rescale_features = [f for f in rescale_features]

                # multi index has problems in the direct assignent so we need to copy back column by column
                tmp = dff[rescale_features].apply(lambda row: ReScaler((row.min(), row.max()), target_range)(row),
                                                  raw=True, result_type='broadcast')
                for col in tmp.columns:
                    dff[col] = tmp[col]

        _log.info(f" make features ... done in {pc() - start_pc: .2f} sec!")

        # finally patch the "values" property for features data frame and return
        dff.__class__ = _RNNShapedValuesDataFrame
        return dff

    @property
    def feature_names(self) -> np.ndarray:
        return np.array(self._features)

    def label_names(self, level_above=None) -> List[Union[Tuple[str, ...],str]]:
        labels = self._labels.encoded_labels_columns \
            if isinstance(self._labels, TargetLabelEncoder) else self._labels

        if isinstance(labels, dict):
            label_columns = []
            for target, target_labels in labels.items():
                for label in (target_labels.encoded_labels_columns if isinstance(target_labels, TargetLabelEncoder) else target_labels):
                    label_columns.append((target, label) if level_above is None else (target, level_above, label))

            return label_columns
        else:
            return labels if level_above is None else [(level_above, col) for col in labels]

    @property
    def labels_df(self) -> pd.DataFrame:
        # here we can do all sorts of tricks and encodings ...
        # joined_kwargs(self._features_and_labels.kwargs, self.)
        df = self._encoder(self._df[self._labels_columns], **self._joined_kwargs).dropna().copy()
        return df if self._label_type is None else df.astype(self._label_type)

    @property
    def source_df(self):
        df = self._df.copy()
        df.columns = pd.MultiIndex.from_product([[SOURCE_COLUMN_NAME], df.columns])
        return df

    @property
    def gross_loss_df(self):
        df = None

        if self._gross_loss is not None:
            labels = self._labels
            for target in (labels.keys() if isinstance(labels, dict) else [None]):
                dfl = self.__call_dynamic(self._gross_loss, self._df, target)
                if isinstance(dfl, pd.Series):
                    if dfl.name is None:
                        dfl.name = target or GROSS_LOSS_COLUMN_NAME
                    dfl = dfl.to_frame()

                dfl.columns = [(GROSS_LOSS_COLUMN_NAME, col) if target is None else (target, GROSS_LOSS_COLUMN_NAME, col)
                               for col in dfl.columns]

                df = dfl if df is None else df.join(dfl)

            # multi level index
            df.columns = pd.MultiIndex.from_tuples(df.columns)

        return df

    @property
    def target_df(self):
        df = None

        if self._targets is not None:
            labels = self._labels
            for i, target in enumerate(labels.keys() if isinstance(labels, dict) else [None]):
                dft = self.__call_dynamic(self._targets, self._df, target)

                if isinstance(dft, pd.Series):
                    if dft.name is None:
                        dft.name = target or TARGET_COLUMN_NAME
                    dft = dft.to_frame()
                elif not isinstance(dft, (pd.Series, pd.DataFrame)):
                    dft = pd.DataFrame({target or TARGET_COLUMN_NAME: dft}, index=self._df.index)

                dft.columns = [(TARGET_COLUMN_NAME, col) if target is None else (target, TARGET_COLUMN_NAME, col)
                               for col in dft.columns]

                df = dft if df is None else df.join(dft)

            # multi level index
            df.columns = pd.MultiIndex.from_tuples(df.columns)

        return df

    def _fix_shape(self, df_features):
        # features eventually are in [feature, row, time_step]
        # but need to be in RNN shape which is [row, time_step, feature]
        feature_arr = df_features.values if self._features_and_labels.feature_lags is None else \
            np.array([df_features[cols].values for cols in self.feature_names], ndmin=3).swapaxes(0, 1).swapaxes(1, 2)

        if len(feature_arr) <= 0:
            _log.warning("empty feature array!")

        return feature_arr

    def __str__(self):
        return f'min required data = {self.min_required_samples}'


class _RNNShapedValuesDataFrame(pd.DataFrame):

    class Loc():
        def __init__(self, df):
            self.df = df

        def __getitem__(self, item):
            res = self.df.loc[item]
            res.__class__ = _RNNShapedValuesDataFrame
            return res

    @property
    def loc(self):
        return _RNNShapedValuesDataFrame.Loc(super(pd.DataFrame, self))

    @property
    def values(self):
        top_level_columns = unique_top_level_columns(self)

        # we need to do a sneaky trick here to get a proper "super" object as super() does not work as expected
        # so we simply rename with an empty dict
        df = self.rename({})

        # features eventually are in [feature, row, time_step]
        # but need to be in RNN shape which is [row, time_step, feature]
        feature_arr = df.values if top_level_columns is None else \
            np.array([df[feature].values for feature in top_level_columns],
                     ndmin=3).swapaxes(0, 1).swapaxes(1, 2)

        if len(feature_arr) <= 0:
            _log.warning("empty feature array!")

        return feature_arr



import logging
import unittest
from typing import List

import pandas as pd
import numpy as np
from sklearn.neural_network import MLPClassifier

import pandas_ml_utils as pdu
from pandas_ml_utils.constants import *
from test.config import TEST_FILE

from pandas_ml_utils.model.features_and_labels.target_encoder import TargetLabelEncoder
from test.mocks.mock_model import MockModel

from pandas_ml_utils.utils.functions import integrate_nested_arrays

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class EncoderTest(unittest.TestCase):

    def test__2d_encoding(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date')
        df["label"] = df["spy_Close"] > df["spy_Open"]

        class ArrayEncoder(TargetLabelEncoder):

            def __init__(self):
                super().__init__()

            @property
            def labels_source_columns(self) -> List[str]:
                return ["spy_Close"]

            @property
            def encoded_labels_columns(self) -> List[str]:
                return ["2D"]

            def encode(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
                res = pd.DataFrame({}, index=df.index)
                res["2D"] = df["spy_Close"] = df["spy_Close"].apply(lambda r: np.array([r, r]))
                return res

        """when"""
        model = MockModel(pdu.FeaturesAndLabels(["spy_Close"], ArrayEncoder(), feature_lags=[0, 1, 2]))
        fit = df.fit(model)

        """then"""
        print(fit.test_summary.df)
        self.assertEqual(fit.test_summary.df.shape, (2682, 2))
        self.assertEqual(integrate_nested_arrays(fit.test_summary.df.values).shape, (2682, 2, 2))



import logging
import unittest

import pandas as pd
from sklearn.neural_network import MLPClassifier

import pandas_ml_utils as pdu
from pandas_ml_utils.constants import *
from test.config import TEST_FILE

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class FeatureAndLabelsExtractorTest(unittest.TestCase):

    def test_extractor(self):
        """given"""
        df = pd.read_csv(TEST_FILE, index_col='Date')

        """when"""
        extractor = df.features_and_label_extractor(
            pdu.SkModel(
                MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001, random_state=42),
                pdu.FeaturesAndLabels(features=['feature'], labels=['label'],
                                      gross_loss=lambda df: df["spy_Close"] - df["spy_Open"],
                                      targets=lambda df: df["spy_Close"],
                                      pre_processor=lambda _df: pdu.LazyDataFrame(
                                          _df,
                                          feature=lambda f: f["vix_Close"].rolling(2).mean(),
                                          label=lambda f: (f["spy_Close"].shift(1) > f["spy_Open"]).shift(-1)))))

        fnl = extractor.features_labels_weights_df

        """then"""
        self.assertListEqual(extractor.features_df.columns.tolist(), ['feature'])
        self.assertListEqual(extractor.labels_df.columns.tolist(), ['label'])
        self.assertListEqual(extractor.source_df[SOURCE_COLUMN_NAME].columns.tolist(), [*df.columns.tolist(), 'feature', 'label'])
        self.assertListEqual(extractor.target_df.columns.tolist(), [(TARGET_COLUMN_NAME, "spy_Close")])
        self.assertListEqual(extractor.gross_loss_df.columns.tolist(), [(GROSS_LOSS_COLUMN_NAME, GROSS_LOSS_COLUMN_NAME)])
        self.assertEqual(fnl[0].shape, (6704, 1))
        self.assertEqual(fnl[1].shape, (6704, 1))
        self.assertIsNone(fnl[2])




from __future__ import annotations

import contextlib
import logging
import os
import tempfile
import uuid
from copy import deepcopy
from typing import List, Callable, TYPE_CHECKING, Tuple, Dict, Any

import dill as pickle
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from pandas_ml_utils.model.features_and_labels.features_and_labels import FeaturesAndLabels
from pandas_ml_utils.model.features_and_labels.target_encoder import TargetLabelEncoder
from pandas_ml_utils.summary.summary import Summary
from pandas_ml_utils.utils.functions import suitable_kwargs, join_kwargs

_log = logging.getLogger(__name__)


class Model(object):
    """
    Represents a statistical or ML model and holds the necessary information how to interpret the columns of a
    pandas *DataFrame* ( :class:`.FeaturesAndLabels` ). Currently available implementations are:
     * SkitModel - provide any skit learn classifier or regressor
     * KerasModel - provide a function returning a compiled keras model
     * MultiModel - provide a model which will copied (and fitted) for each provided target
    """

    @staticmethod
    def load(filename: str):
        """
        Loads a previously saved model from disk. By default `dill <https://pypi.org/project/dill/>`_ is used to
        serialize / deserialize a model.

        :param filename: filename of the serialized model inclusive file extension
        :return: returns a deserialized model
        """
        with open(filename, 'rb') as file:
            model = pickle.load(file)
            if isinstance(model, Model):
                return model
            else:
                raise ValueError("Deserialized pickle was not a Model!")

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[pd.DataFrame], Summary] = Summary,
                 **kwargs):
        """
        All implementations of `Model` need to pass two arguments to `super().__init()__`.

        :param features_and_labels: the :class:`.FeaturesAndLabels` object defining all the features,
                                    feature engineerings and labels
        :param summary_provider: a summary provider in the most simple case just holds a `pd.DataFrame` containing all
                                 the labels and all the predictions and optionally loss and target values. Since
                                 constructors as callables as well it is usually enoug tho just pass the type i.e.
                                 `summary_provider=BinaryClassificationSummary`
        :param kwargs:
        """
        self._features_and_labels = features_and_labels
        self._summary_provider = summary_provider
        self.kwargs = kwargs

    @property
    def features_and_labels(self):
        return self._features_and_labels

    @property
    def summary_provider(self):
        return self._summary_provider

    def plot_loss(self):
        pass

    def __getitem__(self, item):
        """
        returns arguments which are stored in the kwargs filed. By providing a tuple, a default in case of missing
        key can be specified
        :param item: name of the item im the kwargs dict or tuple of name, default
        :return: item or default
        """
        if isinstance(item, tuple) and len(item) == 2:
            return self.kwargs[item[0]] if item[0] in self.kwargs else item[1]
        else:
            return self.kwargs[item] if item in self.kwargs else None

    def save(self, filename: str):
        """
        save model to disk
        :param filename: filename inclusive file extension
        :return: None
        """
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

        print(f"saved model to: {os.path.abspath(filename)}")

    def fit(self,
            x: np.ndarray, y: np.ndarray,
            x_val: np.ndarray, y_val: np.ndarray,
            sample_weight_train: np.ndarray, sample_weight_test: np.ndarray) -> float:
        """
        function called to fit the model
        :param x: x
        :param y: y
        :param x_val: x validation
        :param y_val: y validation
        :param sample_weight_train: sample weights for loss penalisation (default np.ones)
        :param sample_weight_test: sample weights for loss penalisation (default np.ones)
        :return: loss of the fit
        """
        pass

    def predict(self, x: np.ndarray) -> np.ndarray:
        """
        prediction of the model for each target

        :param x: x
        :return: prediction of the model for each target
        """

        pass

    def __call__(self, *args, **kwargs):
        """
        returns a copy pf the model with eventually different configuration (kwargs). This is useful for hyper paramter
        tuning or for MultiModels

        :param args:
        :param kwargs: arguments which are eventually provided by hyperopt or by different targets
        :return:
        """
        if not kwargs:
            return deepcopy(self)
        else:
            raise ValueError(f"construction of model with new parameters is not supported\n{type(self)}: {kwargs}")


class SkModel(Model):

    def __init__(self,
                 skit_model,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[pd.DataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.skit_model = skit_model

    def fit(self,
            x: np.ndarray, y: np.ndarray,
            x_val: np.ndarray, y_val: np.ndarray,
            sample_weight_train: np.ndarray, sample_weight_test: np.ndarray) -> float:
        # shape correction if needed
        y = y.ravel() if len(y.shape) > 1 and y.shape[1] == 1 else y

        # remember fitted model
        self.skit_model = self.skit_model.fit(SkModel.reshape_rnn_as_ar(x), y)

        if getattr(self.skit_model, 'loss_', None):
            return self.skit_model.loss_
        else:
            prediction = self.predict(x)
            if isinstance(self.skit_model, LogisticRegression)\
            or type(self.skit_model).__name__.endswith("Classifier")\
            or type(self.skit_model).__name__.endswith("SVC"):
                from sklearn.metrics import log_loss
                try:
                    return log_loss(prediction > 0.5, y).mean()
                except ValueError as e:
                    if "contains only one label" in str(e):
                        return -100
                    else:
                        raise e
            else:
                from sklearn.metrics import mean_squared_error
                return mean_squared_error(prediction, y).mean()

    def predict(self, x) -> np.ndarray:
        if callable(getattr(self.skit_model, 'predict_proba', None)):
            y_hat = self.skit_model.predict_proba(SkModel.reshape_rnn_as_ar(x))
            return y_hat[:, 1] if len(self.features_and_labels.labels) == 1 and y_hat.shape[1] == 2 else y_hat
        else:
            return self.skit_model.predict(SkModel.reshape_rnn_as_ar(x))

    def plot_loss(self):
        loss_curve = getattr(self.skit_model, 'loss_curve_', None)

        if loss_curve is not None:
            import matplotlib.pyplot as plt
            plt.plot(loss_curve)
        else:
            print("no loss curve found")

    def __str__(self):
        return f'{__name__}({repr(self.skit_model)}, {self.features_and_labels})'

    def __call__(self, *args, **kwargs):
        if not kwargs:
            return deepcopy(self)
        else:
            new_model = SkModel(type(self.skit_model)(**kwargs), self.features_and_labels, self.summary_provider)
            new_model.kwargs = deepcopy(self.kwargs)
            return new_model

    @staticmethod
    def reshape_rnn_as_ar(arr3d):
        if len(arr3d.shape) < 3:
            print("Data was not in RNN shape")
            return arr3d
        else:
            return arr3d.reshape(arr3d.shape[0], arr3d.shape[1] * arr3d.shape[2])


class KerasModel(Model):
    # eventually we need to save and load the weights of the keras model individually by using `__getstate__`
    #  `__setstate__` like described here: http://zachmoshe.com/2017/04/03/pickling-keras-models.html
    if TYPE_CHECKING:
        from keras.models import Model as KModel

    def __init__(self,
                 keras_compiled_model_provider: Callable[[], KModel],
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[pd.DataFrame], Summary] = Summary,
                 epochs: int = 100,
                 callbacks: List[Callable] = [],
                 **kwargs):
        """
        Keras compatible implementation of :class:`.Model`.
        :param keras_compiled_model_provider: a callable which provides an eventually compiled ready to fit keras model.
               if the model is not compiled you ned to pass "optimizer" argument as kwargs.
               NOTE: the tensorflow backend is currently limited to 1.*
        :param features_and_labels: see :class:`.Model`
        :param summary_provider: :class:`.Model`.
        :param epochs: number of epochs passed to the keras fit function
        :param callbacks: a list of callable's providing a keras compatible callback. It is neccessary to pass the
                          callable i.e. BaseLogger or lambda: BaseLogger(...) vs. BaseLogger(...)
        :param kwargs: a list of arguments passed to the keras_compiled_model_provider and keras fit function as they
                       match the signature
        """
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.keras_model_provider = keras_compiled_model_provider
        self.custom_objects = {}

        import keras
        if keras.backend.backend() == 'tensorflow':
            import tensorflow as tf
            self.is_tensorflow = True
            self.graph = tf.Graph()
            with self.graph.as_default():
                self.session = tf.Session(graph=self.graph)
        else:
            self.is_tensorflow = False

        # create keras model
        provider_args = suitable_kwargs(keras_compiled_model_provider, **kwargs)
        keras_model = self._exec_within_session(keras_compiled_model_provider, **provider_args)

        if isinstance(keras_model, Tuple):
            # store custom objects
            for i in range(1, len(keras_model)):
                if hasattr(keras_model[i], "__name__"):
                    self.custom_objects[keras_model[i].__name__] = keras_model[i]
                else:
                    raise ValueError("keras custom object must have a __name__")

            keras_model = keras_model[0]
            _log.warning(f"keras model using custom objects {self.custom_objects}")

        # eventually compile keras model
        if not keras_model.optimizer:
            compile_args = suitable_kwargs(keras_model.compile, **kwargs)
            self._exec_within_session(keras_model.compile, **compile_args)

        # set all members
        self.keras_model = keras_model
        self.epochs = epochs
        self.callbacks = callbacks
        self.history = None

    def fit(self,
            x: np.ndarray, y: np.ndarray,
            x_val: np.ndarray, y_val: np.ndarray,
            sample_weight_train: np.ndarray, sample_weight_test: np.ndarray) -> float:
        fitter_args = suitable_kwargs(self.keras_model.fit, **self.kwargs)

        if "verbose" in self.kwargs and self.kwargs["verbose"] > 0:
            print(f'pass args to fit: {fitter_args}')

        fit_history = self._exec_within_session(self.keras_model.fit,
                                                x, y,
                                                sample_weight=sample_weight_train,
                                                epochs=self.epochs,
                                                validation_data=(x_val, y_val),
                                                callbacks=[cb() for cb in self.callbacks],
                                                **fitter_args)
        if self.history is None:
            self.history = fit_history.history
        else:
            for metric, _ in self.history.items():
                self.history[metric] = self.history[metric] + fit_history.history[metric]

        return min(fit_history.history['loss'])

    def predict(self, x):
        return self._exec_within_session(self.keras_model.predict, x)

    def get_weights(self):
        return self._exec_within_session(self.keras_model.get_weights)

    def set_weights(self, weights):
        self._exec_within_session(self.keras_model.set_weights, weights)

    def plot_loss(self):
        import matplotlib.pyplot as plt

        plt.plot(self.history['val_loss'])
        plt.plot(self.history['loss'])

    def _exec_within_session(self, func, *args, **kwargs):
        if self.is_tensorflow:
            with self.graph.as_default():
                with self.session.as_default():
                    return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains all our instance attributes.
        # Always use the dict.copy() method to avoid modifying the original state.
        state = self.__dict__.copy()

        # remove un-pickleable fields
        if self.is_tensorflow:
            del state['graph']
            del state['session']

        # special treatment for the keras model
        tmp_keras_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        self._exec_within_session(lambda: self.keras_model.save(tmp_keras_file, True, True))
        with open(tmp_keras_file, mode='rb') as file:
            state['keras_model'] = file.read()

        # clean up temp file
        with contextlib.suppress(OSError):
            os.remove(tmp_keras_file)

        # return state
        return state

    def __setstate__(self, state):
        from keras.models import load_model

        # restore keras file
        tmp_keras_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        with open(tmp_keras_file, mode='wb') as file:
            file.write(state['keras_model'])
            del state['keras_model']

        # Restore instance attributes
        self.__dict__.update(state)

        # restore keras model and tensorflow session if needed
        if self.is_tensorflow:
            from keras import backend as K
            import tensorflow as tf
            self.graph = tf.Graph()
            with self.graph.as_default():
                self.session = tf.Session(graph=self.graph)
                K.set_session(self.session)
                self.keras_model = load_model(tmp_keras_file, custom_objects=self.custom_objects)
        else:
            self.keras_model = load_model(tmp_keras_file, custom_objects=self.custom_objects)

        # clean up temp file
        with contextlib.suppress(OSError):
            os.remove(tmp_keras_file)

    def __del__(self):
        if self.is_tensorflow:
            try:
                self.session.close()
            except AttributeError:
                pass

    def __call__(self, *args, **kwargs):
        new_model = KerasModel(self.keras_model_provider,
                               self.features_and_labels,
                               self.summary_provider,
                               self.epochs,
                               deepcopy(self.callbacks),
                               **join_kwargs(deepcopy(self.kwargs), kwargs))

        # copy weights before return
        new_model.set_weights(self.get_weights())
        return new_model


class MultiModel(Model):

    def __init__(self,
                 model_provider: Model,
                 summary_provider: Callable[[pd.DataFrame], Summary] = Summary,
                 loss_alpha: float = 0.5,
                 target_kwargs: Dict[str, Dict[str, Any]] = None,
                 **kwargs: Dict):
        assert isinstance(model_provider.features_and_labels.labels, (TargetLabelEncoder, Dict))
        super().__init__(
            # if we have a target args and a target encoder then we need generate multiple targets with different kwargs
            model_provider.features_and_labels.with_labels(
                {target: model_provider.features_and_labels.labels.with_kwargs(**kwargs) for target, kwargs in target_kwargs.items()}) \
                    if target_kwargs and isinstance(model_provider.features_and_labels.labels, TargetLabelEncoder) \
             else model_provider.features_and_labels,
            summary_provider,
            **join_kwargs(model_provider.kwargs, kwargs))

        if isinstance(model_provider, MultiModel):
            raise ValueError("Nesting Multi Models is not supported, you might use a flat structure of all your models")

        self.models = {target: model_provider() for target in self.features_and_labels.labels.keys()}
        self.model_provider = model_provider
        self.target_kwargs = target_kwargs
        self.loss_alpha = loss_alpha

    def fit(self,
            x: np.ndarray, y: np.ndarray,
            x_val: np.ndarray, y_val: np.ndarray,
            sample_weight_train: np.ndarray, sample_weight_test: np.ndarray) -> float:
        losses = []
        pos = 0

        for target, labels in self.features_and_labels.labels.items():
            index = range(pos, pos + len(labels))
            target_y = y[:,index]
            target_y_val = y_val[:,index]
            target_w = sample_weight_train[:,index] if sample_weight_train is not None else None
            target_w_val = sample_weight_test[:,index] if sample_weight_test is not None else None
            _log.info(f"fit model for target {target}")
            losses.append(self.models[target].fit(x, target_y, x_val, target_y_val, target_w, target_w_val))
            pos += len(labels)

        losses = np.array(losses)
        a = self.loss_alpha

        # return weighted loss between mean and max loss
        return (losses.mean() * (1 - a) + a * losses.max()) if len(losses) > 0 else None

    def predict(self, x: np.ndarray) -> np.ndarray:
        # intermediate function to concatenate individual predictions
        def predict_as_column_matrix(target):
            prediction = self.models[target].predict(x)

            # eventually fix shape to have 2 dimensions
            if len(prediction.shape) <= 1:
                prediction = prediction.reshape((-1, 1))

            # fix dimensions if prediction length is 2 and expected length is 1
            if prediction.shape[1] == 2 and len(self.features_and_labels.labels[target]) == 1:
                prediction = prediction[:, 1].reshape((-1, 1))

            # return prediction with expected shape
            return prediction

        # return all the concatenated predictions
        return np.concatenate([predict_as_column_matrix(target) for target in self.features_and_labels.labels.keys()],
                              axis=1)

    def __call__(self, *args, **kwargs):
        new_multi_model = MultiModel(self.model_provider, self.summary_provider, self.loss_alpha, self.target_kwargs)

        if kwargs:
            raise ValueError("kwargs on cloning multi model ist currently not supported!")

        return new_multi_model



## THIS need to be fixed somewhen
#class OpenAiGymModel(Model):
#    from typing import TYPE_CHECKING
#    if TYPE_CHECKING:
#        from rl.core import Agent
#
#    def __init__(self,
#                 agent_provider: Callable[[Tuple, int], Agent],
#                 features_and_labels: FeaturesAndLabels,
#                 action_reward_functions: List[Callable[[np.ndarray], float]],
#                 reward_range: Tuple[int, int],
#                 oservation_range: Tuple[int, int] = None,
#                 episodes: int = 1000,
#                 **kwargs):
#        super().__init__(features_and_labels, **kwargs)
#        self.agent_provider = agent_provider
#        self.action_reward_functions = action_reward_functions
#        self.reward_range = reward_range
#        self.oservation_range = oservation_range
#        self.episodes = episodes
#        self.agent = agent_provider(features_and_labels.shape()[0], len(action_reward_functions))
#
#        # some history
#        self.keras_train_history = None
#        self.keras_test_history = None
#        self.history = ()
#
#    def fit(self, x, y, x_val, y_val, df_index_train, df_index_test):
#        mm = (min([x.min(), x_val.min()]), max([x.max(), x_val.max()])) if self.oservation_range is None else self.oservation_range
#        training_gym = RowWiseGym((df_index_train, x, y), self.features_and_labels, self.action_reward_functions, self.reward_range, mm)
#        test_gym = RowWiseGym((df_index_test, x_val, y_val), self.features_and_labels, self.action_reward_functions, self.reward_range, mm)
#
#        self.keras_train_history = self.agent.fit(training_gym, nb_steps=len(x) * self.episodes)
#        #self.keras_test_history = self.agent.test(test_gym, nb_episodes=1) # clarification needed what test actually does: https://github.com/keras-rl/keras-rl/issues/342
#        test_gym = self._forward_gym(test_gym)
#
#        self.history = (training_gym.get_history(), test_gym.get_history())
#
#    def back_test(self, index, x, y):
#        mm = (x.min(), x.max()) if self.oservation_range is None else self.oservation_range
#        gym = RowWiseGym((index, x, y), self.features_and_labels, self.action_reward_functions, self.reward_range, mm)
#        return self._forward_gym(gym).get_history()
#
#    def _predict(self, x: np.ndarray, target: str) -> np.ndarray:
#        return [self.agent.forward(x[r]) for r in range(len(x))]
#
#    def __call__(self, *args, **kwargs):
#        if kwargs:
#            raise ValueError("hyper parameter tunig currently not supported for RL")
#
#        return OpenAiGymModel(self.agent_provider,
#                              self.features_and_labels,
#                              self.action_reward_functions,
#                              self.reward_range,
#                              self.episodes,
#                              **deepcopy(self.kwargs))
#
#    def _forward_gym(self, gym):
#        done = False
#        state = gym.reset()
#        while not done:
#            state, reward, done, _ = gym.step(self.agent.forward(state))
#
#        return gym
#
#class StableBaselineModel(Model):
#    # add stable baseline models https://stable-baselines.readthedocs.io/en/master/guide/algos.html
#    pass


from unittest import TestCase

import numpy as np

from pandas_ml_utils.sampling.boosting import KFoldBoostRareEvents, KEquallyWeightEvents
from pandas_ml_utils.utils.classes import ReScaler
from pandas_ml_utils.utils.functions import unfold_parameter_space, one_hot


class TestUtilClasses(TestCase):

    def test_rescaler(self):
        """given"""
        data = np.random.random(50)
        scaler = ReScaler((data.min(), data.max()), (1, 2))

        """when"""
        scaled = scaler(data)

        """then"""
        np.testing.assert_almost_equal(scaled.min(), 1)
        np.testing.assert_almost_equal(scaled.max(), 2)
        self.assertEqual(np.argmax(data), np.argmax(scaled))

    def test_rescaler_reverse(self):
        """given"""
        data = np.random.random(50)
        scaler = ReScaler((data.min(), data.max()), (2, 1))

        """when"""
        scaled = scaler(data)

        """then"""
        np.testing.assert_almost_equal(scaled.min(), 1)
        np.testing.assert_almost_equal(scaled.max(), 2)
        self.assertNotEqual(np.argmax(scaled), len(data) - 1 - np.argmax(data))